from collections import defaultdict
from textwrap import indent

from ..parser import ast
from .utils import INTEGER_TYPES
from .utils import METHOD_BIN_OPERATORS
from .utils import CompileError
from .utils import GenericType
from .utils import format_mys_type
from .utils import get_import_from_info
from .utils import has_docstring
from .utils import is_pascal_case
from .utils import is_snake_case
from .utils import is_upper_snake_case
from .utils import make_function_name


class TypeVisitor(ast.NodeVisitor):

    def visit_Name(self, node):
        return node.id

    def visit_List(self, node):
        nitems = len(node.elts)

        if nitems != 1:
            raise CompileError(f"expected 1 type in list, got {nitems}", node)

        return [self.visit(elem) for elem in node.elts]

    def visit_Tuple(self, node):
        return tuple(self.visit(elem) for elem in node.elts)

    def visit_Dict(self, node):
        return {node.keys[0].id: self.visit(node.values[0])}

    def visit_Set(self, node):
        nitems = len(node.elts)

        if nitems != 1:
            raise CompileError(f"expected 1 type in set, got {nitems}", node)

        return {self.visit(node.elts[0])}

    def visit_Subscript(self, node):
        types = self.visit(node.slice)

        if isinstance(node.slice, ast.Name):
            types = [types]
        else:
            types = list(types)

        return GenericType(self.visit(node.value), types, node)


class FormatDefaultVisitor(ast.NodeVisitor):

    def visit_Constant(self, node):
        if isinstance(node.value, str):
            return f'"{node.value}"'
        elif isinstance(node.value, tuple) and len(node.value) == 3:
            return f"'{node.value[0]}'"
        else:
            return str(node.value)

    def visit_UnaryOp(self, node):
        operand = self.visit(node.operand)

        if isinstance(node.op, ast.USub):
            return f'-{operand}'
        else:
            return operand

    def visit_List(self, node):
        return '[' + ', '.join([self.visit(elem) for elem in node.elts]) + ']'

    def visit_Tuple(self, node):
        return '(' + ', '.join([self.visit(elem) for elem in node.elts]) + ')'

    def visit_Dict(self, node):
        return '{' + ', '.join([
            f'{self.visit(key)}: {self.visit(value)}'
            for key, value in zip(node.keys, node.values)
        ]) + '}'

    def visit_Set(self, node):
        return '{' + ', '.join([self.visit(elem) for elem in node.elts]) + '}'

    def visit_Call(self, node):
        params = ', '.join([self.visit(arg) for arg in node.args])

        return f'{node.func.id}({params})'

    def visit_Name(self, node):
        return node.id

    def visit_Attribute(self, node):
        return f'{self.visit(node.value)}.{node.attr}'


class Function:

    def __init__(self,
                 name,
                 generic_types,
                 raises,
                 is_test,
                 args,
                 returns,
                 node,
                 module_name=None,
                 is_overloaded=False,
                 docstring=None):
        self.name = name
        self.generic_types = generic_types
        self.raises = raises
        self.is_test = is_test
        self.args = args
        self.returns = returns
        self.node = node
        self.module_name = module_name
        self.is_overloaded = is_overloaded
        self.docstring = docstring

    def is_generic(self):
        return bool(self.generic_types)

    def make_name(self):
        if self.is_overloaded:
            return make_function_name(
                self.name,
                [param.type for param, _ in self.args],
                self.returns)
        else:
            return self.name

    def signature_string(self, method):
        params = []

        for param, default in self.args:
            param = f'{param.name}: {format_mys_type(param.type)}'

            if default is not None:
                param += ' = '
                param += FormatDefaultVisitor().visit(default)

            params.append(param)

        params_string = ', '.join(params)

        if len(params_string) > 60:
            if method:
                if params:
                    params = ['self'] + params
                else:
                    params = ['self']

            params_string = indent(',\n'.join(params),
                                   ' ' * (len(self.name) + 5)).strip()
        else:
            if method:
                if params_string:
                    params_string = 'self, ' + params_string
                else:
                    params_string = 'self'

        if self.returns is None:
            returns = ''
        else:
            returns = f' -> {format_mys_type(self.returns)}'

        return f'{self.name}({params_string}){returns}'

    def __str__(self):
        args = [f'{param.name}: {param.type}' for param, _ in self.args]

        return (
            f'Function(name={self.name}, generic_types={self.generic_types}, '
            f'args={args}, returns={self.returns}, module_name={self.module_name}, '
            f'is_overloaded={self.is_overloaded})')

class Param:

    def __init__(self, name, type_, node):
        self.name = name
        self.type = type_
        self.node = node

    def __str__(self):
        return f'Param(name={self.name}, type={self.type})'

class Member:

    def __init__(self, name, type_, node):
        self.name = name
        self.type = type_
        self.node = node

    def __str__(self):
        return f'Member(name={self.name}, type={self.type})'


class Class:

    def __init__(self,
                 name,
                 generic_types,
                 members,
                 methods,
                 functions,
                 implements,
                 node,
                 module_name=None,
                 docstring=None):
        self.name = name
        self.generic_types = generic_types
        self.members = members
        self.methods = methods
        self.functions = functions
        self.implements = implements
        self.node = node
        self.module_name = module_name
        self.docstring = docstring

    def is_generic(self):
        return bool(self.generic_types)

    def __str__(self):
        members = [str(member) for member in self.members.values()]
        methods = []
        functions = []

        for methods_by_name in self.methods.values():
            for method in methods_by_name:
                methods.append(str(method))

        for functions_by_name in self.functions.values():
            for function in functions_by_name:
                functions.append(str(function))

        return (
            f'Class(name={self.name}, generic_types={self.generic_types}, '
            f'members={members}, methods={methods}, '
            f'functions={functions}, implements={self.implements})')


class Trait:

    def __init__(self, name, methods, node, docstring):
        self.name = name
        self.methods = methods
        self.node = node
        self.docstring = docstring


class Enum:

    def __init__(self, name, type_, members, docstring):
        self.name = name
        self.type = type_
        self.members = members
        self.member_names = {name for name, _ in members}
        self.docstring = docstring


class Variable:

    def __init__(self,
                 name,
                 type_,
                 node,
                 docstring):
        self.name = name
        self.type = type_
        self.node = node
        self.docstring = docstring

    def __str__(self):
        return f'Variable(name={self.name}, type={self.type})'

class Definitions:
    """Defined variables, classes, traits, enums and functions for one
    module. This information is useful when verifying that modules
    uses this module correctly.

    All function and method names are guaranteed to be snake
    case. Same applies to their parameter anmes.

    All classes, traits and enums names are guaranteed to be pascal
    case.

    Variable names can either be lower or upper case snake case.

    """

    def __init__(self, module_name):
        self.module_name = module_name
        self.variables = {}
        self.classes = {}
        self.traits = {}
        self.enums = {}
        self.functions = defaultdict(list)
        self.imports = defaultdict(list)

    def _check_unique_name(self, name, node, is_function=False):
        if name in self.variables:
            raise CompileError(f"there is already a variable called '{name}'", node)

        if name in self.classes:
            raise CompileError(f"there is already a class called '{name}'", node)

        if name in self.traits:
            raise CompileError(f"there is already a trait called '{name}'", node)

        if name in self.enums:
            raise CompileError(f"there is already an enum called '{name}'", node)

        if not is_function:
            if name in self.functions:
                raise CompileError(f"there is already a function called '{name}'",
                                   node)

    def define_variable(self, name, value, node):
        self._check_unique_name(name, node)
        self.variables[name] = value

    def define_class(self, name, value, node):
        self._check_unique_name(name, node)
        value.module_name = self.module_name
        self.classes[name] = value

    def define_trait(self, name, value, node):
        self._check_unique_name(name, node)
        self.traits[name] = value

    def define_enum(self, name, value, node):
        self._check_unique_name(name, node)
        self.enums[name] = value

    def define_function(self, name, value, node):
        self._check_unique_name(name, node, True)
        value.module_name = self.module_name
        functions = self.functions[name]

        if functions:
            value.is_overloaded = True
            functions[0].is_overloaded = True

        functions.append(value)

    def add_import(self, module, name, asname):
        self.imports[asname].append((module, name))

    def __str__(self):
        result = ['Definitions:']

        result.append('  Variables:')

        for definition in self.variables.values():
            result.append(f'    {definition.name}: {definition.type}')

        result.append('  Classes:')

        for definition in self.classes.values():
            bases = ', '.join(definition.implements)
            result.append(f'    {definition.name}({bases}):')

            for member in definition.members.values():
                result.append(f'      {member.name}: {member.type}')

            for methods in definition.methods.values():
                for method in methods:
                    params = ', '.join([
                        f'{param.name}: {param.type}'
                        for param, _ in method.args
                    ])
                    result.append(
                        f'      {method.name}({params}) -> {method.returns}')

        result.append('  Traits:')

        for definition in self.traits.values():
            result.append(f'    {definition.name}:')

            for methods in definition.methods.values():
                for method in methods:
                    params = ', '.join([
                        f'{param.name}: {param.type}'
                        for param, _ in method.args
                    ])
                    result.append(
                        f'      {method.name}({params}) -> {method.returns}')

        result.append('  Enums:')

        for enum in self.enums:
            result.append(f'    {enum}')

        result.append('  Functions:')

        for functions in self.functions.values():
            for function in functions:
                params = ', '.join([
                    f'{param.name}: {param.type}'
                    for param, _ in function.args
                ])
                result.append(
                    f'    {function.name}({params}) -> {function.returns}')

        return '\n'.join(result)


def is_method(node):
    return len(node.args) >= 1 and node.args[0].arg == 'self'


class FunctionVisitor(TypeVisitor):

    ALLOWED_DECORATORS = ['generic', 'test', 'raises']

    def visit_arg(self, node):
        if node.annotation is None:
            raise CompileError("parameters must have a type", node)

        if not is_snake_case(node.arg):
            raise CompileError("parameter names must be snake case", node)

        return Param(node.arg, self.visit(node.annotation), node)

    def visit_arguments(self, node):
        args = []

        for i, arg in enumerate(node.args[::-1]):
            if i < len(node.defaults):
                args.append((self.visit(arg),
                             node.defaults[len(node.defaults) - i - 1]))
            else:
                args.append((self.visit(arg), None))

        return args[::-1]

    def visit_FunctionDef(self, node):
        if not is_snake_case(node.name):
            raise CompileError("function names must be snake case", node)

        decorators = visit_decorator_list(node.decorator_list,
                                          self.ALLOWED_DECORATORS)
        args = self.visit(node.args)

        if node.returns is None:
            returns = None
        else:
            returns = FunctionVisitor().visit(node.returns)

        if has_docstring(node):
            docstring = node.body[0].value.value
        else:
            docstring = None

        return Function(node.name,
                        decorators.get('generic', []),
                        decorators.get('raises', []),
                        'test' in decorators,
                        args,
                        returns,
                        node,
                        None,
                        None,
                        docstring)


class MethodVisitor(FunctionVisitor):

    ALLOWED_DECORATORS = ['generic', 'raises']

    def visit_arguments(self, node):
        args = []

        for i, arg in enumerate(node.args[1:][::-1]):
            if i < len(node.defaults):
                args.append((self.visit(arg),
                             node.defaults[len(node.defaults) - i - 1]))
            else:
                args.append((self.visit(arg), None))

        return args[::-1]


def visit_decorator_list(decorator_list, allowed_decorators):
    decorators = {}

    for decorator in decorator_list:
        if isinstance(decorator, ast.Call):
            name = decorator.func.id
            values = []

            for arg in decorator.args:
                if not isinstance(arg, ast.Name):
                    raise CompileError("invalid decorator value", arg)

                if arg.id in values:
                    raise CompileError(f"'{arg.id}' can only be given once", arg)

                values.append(arg.id)
        elif isinstance(decorator, ast.Name):
            name = decorator.id
            values = []

            if name == 'enum':
                values.append('i64')
        else:
            raise CompileError("decorators must be @name or @name()", decorator)

        if name not in allowed_decorators:
            raise CompileError(f"invalid decorator '{name}'", decorator)

        if name in decorators:
            raise CompileError(f"@{name} can only be given once", decorator)

        if name == 'enum':
            if len(values) != 1:
                raise CompileError(f"one parameter expected, got {len(values)}",
                                   decorator)

            if values[0] not in INTEGER_TYPES:
                raise CompileError(f"integer type expected, not '{values[0]}'",
                                   decorator.args[0])

            decorators['enum'] = values[0]
        elif name == 'trait':
            if values:
                raise CompileError("no parameters expected", decorator)

            decorators['trait'] = None
        elif name == 'test':
            if values:
                raise CompileError("no parameters expected", decorator)

            decorators['test'] = None
        elif name == 'generic':
            if not values:
                raise CompileError("at least one parameter required", decorator)

            decorators['generic'] = values
        elif name == 'raises':
            if not values:
                raise CompileError("@raises requires at least one error", decorator)

            decorators['raises'] = values

    return decorators


def validate_method_signature(method, method_node):
    if method.name in METHOD_BIN_OPERATORS:
        if len(method.args) != 1:
            raise CompileError(
                f'{method.name} must take exactly one parameter', method_node)

        if method.returns is None:
            raise CompileError(f'{method.name} must return a value', method_node)


class DefinitionsVisitor(ast.NodeVisitor):

    def __init__(self, source_lines, module_levels, module_name):
        super().__init__()
        self._source_lines = source_lines
        self._module_levels = module_levels
        self._definitions = Definitions(module_name)
        self._next_enum_value = None

    def visit_Module(self, node):
        body_iter = iter(node.body)
        item_index = 0

        for item in body_iter:
            if isinstance(item, ast.AnnAssign):
                if self.visit_global_variable(item, node.body, item_index + 1):
                    next(body_iter)
                    item_index += 1
            else:
                self.visit(item)

            item_index += 1

        return self._definitions

    def visit_ImportFrom(self, node):
        self._definitions.add_import(*get_import_from_info(node, self._module_levels))

    def visit_Assign(self, node):
        raise CompileError("global variable types cannot be inferred", node)

    def visit_global_variable(self, node, body, next_index):
        docstring = None

        if len(body) > next_index:
            docstring_node = body[next_index]

            if isinstance(docstring_node, ast.Expr):
                if isinstance(docstring_node.value, ast.Constant):
                    if isinstance(docstring_node.value.value, str):
                        docstring = docstring_node.value.value

        name = node.target.id

        if not is_upper_snake_case(name):
            raise CompileError(
                "global variable names must be upper case snake case", node)

        self._definitions.define_variable(
            name,
            Variable(name,
                     TypeVisitor().visit(node.annotation),
                     node,
                     docstring),
            node)

        return docstring is not None

    def visit_enum_member_assign(self, node):
        if len(node.targets) != 1:
            raise CompileError("invalid enum member syntax", node)

        if not isinstance(node.targets[0], ast.Name):
            raise CompileError("invalid enum member syntax", node)

        name = node.targets[0].id
        sign = 1

        if isinstance(node.value, ast.Constant):
            if isinstance(node.value.value, tuple):
                if len(node.value.value) == 1:
                    return (name, node.value.value[0])

        if isinstance(node.value, ast.UnaryOp):
            if isinstance(node.value.op, ast.USub):
                sign = -1
            else:
                raise CompileError("invalid enum member value", node.value)

            value = node.value.operand
        else:
            value = node.value

        if isinstance(value, ast.Constant):
            if not isinstance(value.value, int):
                raise CompileError("invalid enum member value", value)

            value = sign * value.value
        else:
            raise CompileError("invalid enum member value", node.value)

        if self._next_enum_value is not None:
            if value < self._next_enum_value:
                raise CompileError("enum member value lower than for previous member",
                                   node.value)

        self._next_enum_value = (value + 1)

        return (name, value)

    def next_enum_value(self):
        if self._next_enum_value is None:
            self._next_enum_value = 0

        value = self._next_enum_value
        self._next_enum_value += 1

        return value

    def visit_enum_member_expression(self, node):
        if not isinstance(node.value, ast.Name):
            raise CompileError("invalid enum member syntax", node)

        name = node.value.id
        value = self.next_enum_value()

        return (name, value)

    def visit_enum_member(self, node):
        if isinstance(node, ast.Assign):
            name, value = self.visit_enum_member_assign(node)
        elif isinstance(node, ast.Expr):
            name, value = self.visit_enum_member_expression(node)
        else:
            raise CompileError("invalid enum member syntax", node)

        if not is_pascal_case(name):
            raise CompileError("enum member names must be pascal case", node)

        return (name, value)

    def visit_enum(self, node, decorators):
        enum_name = node.name

        if not is_pascal_case(enum_name):
            raise CompileError("enum names must be pascal case", node)

        self._next_enum_value = None
        members = []
        body_iter = iter(node.body)

        if has_docstring(node):
            docstring = node.body[0].value.value
            next(body_iter)
        else:
            docstring = None

        for item in body_iter:
            members.append(self.visit_enum_member(item))

        self._definitions.define_enum(enum_name,
                                      Enum(enum_name,
                                           decorators['enum'],
                                           members,
                                           docstring),
                                      node)

    def visit_trait(self, node):
        trait_name = node.name

        if not is_pascal_case(trait_name):
            raise CompileError("trait names must be pascal case", node)

        methods = defaultdict(list)
        body_iter = iter(node.body)

        if has_docstring(node):
            docstring = node.body[0].value.value
            next(body_iter)
        else:
            docstring = None

        for item in body_iter:
            if isinstance(item, ast.FunctionDef):
                name = item.name

                if is_method(item.args):
                    methods[name].append(MethodVisitor().visit(item))
            elif isinstance(item, ast.AnnAssign):
                raise CompileError('traits cannot have members', item)

        self._definitions.define_trait(trait_name,
                                       Trait(trait_name, methods, node, docstring),
                                       node)

    def visit_class(self, node, decorators):
        class_name = node.name

        if not is_pascal_case(class_name):
            raise CompileError("class names must be pascal case", node)

        methods = defaultdict(list)
        functions = defaultdict(list)
        members = {}

        generic_types = decorators.get('generic', [])
        implements = {
            trait.id: trait
            for trait in node.bases
        }
        body_iter = iter(node.body)

        if has_docstring(node):
            docstring = node.body[0].value.value
            next(body_iter)
        else:
            docstring = None

        for item in body_iter:
            if isinstance(item, ast.FunctionDef):
                name = item.name

                if is_method(item.args):
                    method = MethodVisitor().visit(item)
                    validate_method_signature(method, item)
                    methods[name].append(method)
                else:
                    functions[name].append(FunctionVisitor().visit(item))
            elif isinstance(item, ast.AnnAssign):
                name = item.target.id

                if not is_snake_case(name):
                    raise CompileError("class member names must be snake case", item)

                members[name] = Member(name,
                                       TypeVisitor().visit(item.annotation),
                                       item)

                if item.value is not None:
                    raise CompileError("class members cannot have default values",
                                       item.value)

        self._definitions.define_class(class_name,
                                       Class(class_name,
                                             generic_types,
                                             members,
                                             methods,
                                             functions,
                                             implements,
                                             node,
                                             None,
                                             docstring),
                                       node)

    def visit_ClassDef(self, node):
        decorators = visit_decorator_list(node.decorator_list,
                                          ['enum', 'trait', 'generic'])

        if 'enum' in decorators:
            self.visit_enum(node, decorators)
        elif 'trait' in decorators:
            self.visit_trait(node)
        else:
            self.visit_class(node, decorators)

    def visit_FunctionDef(self, node):
        self._definitions.define_function(node.name,
                                          FunctionVisitor().visit(node),
                                          node)


def find_definitions(tree, source_lines, module_levels, module_name):
    """Find all definitions in given tree and return them.

    """

    return DefinitionsVisitor(source_lines,
                              module_levels,
                              module_name).visit(tree)


class MakeFullyQualifiedNames:

    def __init__(self, module, module_definitions):
        self.module = module
        self.module_definitions = module_definitions

    def process_type(self, mys_type):
        if isinstance(mys_type, list):
            return [self.process_type(mys_type[0])]
        elif isinstance(mys_type, set):
            return {self.process_type(list(mys_type)[0])}
        elif isinstance(mys_type, dict):
            return {
                self.process_type(list(mys_type.keys())[0]):
                self.process_type(list(mys_type.values())[0])
            }
        elif isinstance(mys_type, tuple):
            return tuple(self.process_type(item) for item in mys_type)
        elif isinstance(mys_type, GenericType):
            mys_type.name = self.process_type(mys_type.name)
            types = []

            for type_ in mys_type.types:
                types.append(self.process_type(type_))

            mys_type.types = types

            return mys_type
        elif mys_type in self.module_definitions.classes:
            return f'{self.module}.{mys_type}'
        elif mys_type in self.module_definitions.traits:
            return f'{self.module}.{mys_type}'
        elif mys_type in self.module_definitions.enums:
            return f'{self.module}.{mys_type}'
        elif mys_type in self.module_definitions.imports:
            imported_module, new_type = self.module_definitions.imports[mys_type][0]

            return f'{imported_module}.{new_type}'
        else:
            return mys_type

    def process_function(self, function):
        function.returns = self.process_type(function.returns)

        for param, _ in function.args:
            param.type = self.process_type(param.type)

    def process_variable(self, variable_definition):
        variable_definition.type = self.process_type(variable_definition.type)

    def process_class(self, class_definition):
        for base in list(class_definition.implements):
            node = class_definition.implements.pop(base)

            if base in self.module_definitions.traits:
                base = self.process_type(base)
            elif base in self.module_definitions.imports:
                imported_module, name = self.module_definitions.imports[base][0]
                base = f'{imported_module}.{name}'
            elif base == 'Error':
                pass
            else:
                raise CompileError('trait does not exist', node)

            class_definition.implements[base] = node

        for member in class_definition.members.values():
            member.type = self.process_type(member.type)

        for methods in class_definition.methods.values():
            for method in methods:
                self.process_function(method)

    def process_trait(self, trait_definition):
        for methods in trait_definition.methods.values():
            for method in methods:
                self.process_function(method)

    def process_module(self):
        """Make variable types, members, parameters and return types and
        implemented traits fully qualified names.

        """

        for variable_definition in self.module_definitions.variables.values():
            self.process_variable(variable_definition)

        for class_definition in self.module_definitions.classes.values():
            self.process_class(class_definition)

        for trait_definition in self.module_definitions.traits.values():
            self.process_trait(trait_definition)

        for functions in self.module_definitions.functions.values():
            for function in functions:
                self.process_function(function)

    def process(self):
        return self.process_module()


def make_fully_qualified_names_module(module, module_definitions):
    """Make variable types, members, parameters and return types and
    implemented traits fully qualified names.

    """

    return MakeFullyQualifiedNames(module, module_definitions).process()



import re

from ..parser import ast
from .cpp_reserved import make_cpp_safe_name


class GenericType:

    def __init__(self, name, types, node):
        self.name = name
        self.types = types
        self.node = node

    def __str__(self):
        return f'GenericType(name={self.name}, types={self.types})'


class CompileError(Exception):

    def __init__(self, message, node):
        super().__init__()
        self.message = message
        self.lineno = node.lineno
        self.offset = node.col_offset


class InternalError(CompileError):
    pass


SNAKE_CASE_RE = re.compile(r'^(_*[a-z][a-z0-9_]*)$')
UPPER_SNAKE_CASE_RE = re.compile(r'^(_*[A-Z][A-Z0-9_]*)$')
PASCAL_CASE_RE = re.compile(r'^_?[A-Z][a-zA-Z0-9]*$')

INTEGER_TYPES = set(['i8', 'i16', 'i32', 'i64', 'u8', 'u16', 'u32', 'u64'])
NUMBER_TYPES = INTEGER_TYPES | set(['f32', 'f64'])
PRIMITIVE_TYPES = NUMBER_TYPES | set(['bool', 'char'])
BUILTIN_TYPES = PRIMITIVE_TYPES | set(['string', 'bytes'])

METHOD_BIN_OPERATORS = {
    '__add__',
    '__sub__',
    '__mul__',
    '__div__',
    '__eq__',
    '__ne__',
    '__gt__',
    '__ge__',
    '__lt__',
    '__le__'
}

METHOD_AUG_OPERATORS = {
    '__iadd__',
    '__isub__',
    '__imul__',
    '__idiv__'
}

METHOD_OPERATORS = METHOD_BIN_OPERATORS | METHOD_AUG_OPERATORS

BUILTIN_ERRORS = {
    'KeyError',
    'IndexError',
    'InterruptError',
    'NotImplementedError',
    'AssertionError',
    'SystemExitError',
    'ValueError',
    'UnreachableError'
}

BUILTIN_CALLS = set(
    list(INTEGER_TYPES)
    + list(BUILTIN_ERRORS) + [
        'print',
        'bytes',
        'char',
        'list',
        'input',
        'set',
        'regex',
        'str',
        'min',
        'max',
        'len',
        'abs',
        'f32',
        'f64',
        'enumerate',
        'range',
        'reversed',
        'slice',
        'sum',
        'zip',
        'string',
        'default',
        '__MYS_TRACEBACK_ENTER',
        '__MYS_TRACEBACK_EXIT',
        '__MYS_TRACEBACK_SET'
    ])

OPERATORS = {
    ast.Add: '+',
    ast.Sub: '-',
    ast.Mult: '*',
    ast.Div: '/',
    ast.Mod: '%',
    ast.LShift: '<<',
    ast.RShift: '>>',
    ast.BitOr: '|',
    ast.BitXor: '^',
    ast.BitAnd: '&',
    ast.FloorDiv: '/',
    ast.Not: '!',
    ast.Invert: '~',
    ast.UAdd: '+',
    ast.USub: '-',
    ast.Is: '==',
    ast.Eq: '==',
    ast.NotEq: '!=',
    ast.Lt: '<',
    ast.LtE: '<=',
    ast.Gt: '>',
    ast.GtE: '>='
}

COMPARISON_METHODS = {
    ast.Eq: '__eq__',
    ast.NotEq: '__ne__',
    ast.Lt: '__lt__',
    ast.LtE: '__le__',
    ast.Gt: '__gt__',
    ast.GtE: '__ge__'
}

OPERATORS_TO_METHOD = {
    ast.Add: '__add__',
    ast.Sub: '__sub__',
    ast.Mult: '__mul__',
    ast.Div: '__div__'
}

OPERATORS_TO_AUG_METHOD = {
    ast.Add: '__iadd__',
    ast.Sub: '__isub__',
    ast.Mult: '__imul__',
    ast.Div: '__idiv__'
}

SET_METHODS = {
    'is_disjoint': [['set'], 'bool'],
    'is_subset': [['set'], 'bool'],
    'is_proper_subset': [['set'], 'bool'],
    'is_superset': [['set'], 'bool'],
    'is_proper_superset': [['set'], 'bool'],
    'union': [['set'], 'set'],
    'update': [['set'], 'set'],
    'intersection': [['set'], 'set'],
    'intersection_update': [['set'], 'set'],
    'difference': [['set'], 'set'],
    'difference_update': [['set'], 'set'],
    'symmetric_difference': [['set'], 'set'],
    'symmetric_difference_update': [['set'], 'set'],
    'add': [['elem'], None],
    'clear': [[], None],
    'discard': [['elem'], None],
    'remove': [['elem'], None]
}

STRING_METHODS = {
    'to_utf8': [[], 'bytes'],
    'starts_with': [['string'], 'bool'],
    'ends_with': [['string'], 'bool'],
    'join': [[['string']], 'string'],
    'lower': [[], 'string'],
    'upper': [[], 'string'],
    'casefold': [[], 'string'],
    'capitalize': [[], 'string'],
    'split': [['string'], ['string']],
    'strip': [['string'], 'string'],
    'strip_left': [['string'], 'string'],
    'strip_right': [['string'], 'string'],
    'find': [['string|char', 'optional<i64>', 'optional<i64>'], 'i64'],
    'find_reverse': [['string|char', 'optional<i64>', 'optional<i64>'], 'i64'],
    'partition': [['string'], ('string', 'string', 'string')],
    'replace': [[None, None], 'string'],
    'is_alpha': [[], 'bool'],
    'is_digit': [[], 'bool'],
    'is_numeric': [[], 'bool'],
    'is_space': [[], 'bool'],
    'match': [['regex'], 'regexmatch']
}

BYTES_METHODS = {
    'to_hex': [[], 'string'],
    'find': [['string', 'optional<i64>', 'optional<i64>'], 'i64']
}

LIST_METHODS = {
    'append': [['<listtype>'], None],
    'extend': [['list<listtype>'], None],
    'insert': [['i64', '<listtype>'], None],
    'remove': [['<listtype>'], None],
    'reverse': [[], None],
    'sort': [[], None],
    'count': [['<listtype>'], 'i64'],
    'pop': [['i64'], '<listtype>'],
    'clear': [[], None]
}

REGEX_METHODS = {
    'split': [['string'], ['string']],
    'match': [['string'], 'regexmatch'],
    'replace': [['string', 'string'], 'string']
}

REGEXMATCH_METHODS = {
    'span': [[None], ('i64', 'i64')],
    'start': [[None], 'i64'],
    'end': [[None], 'i64'],
    'group': [[None], 'string'],
    'groups': [[None], ['string']],
    'group_dict': [[], {'string': 'string'}]
}

CHAR_METHODS = {
    'is_alpha': [[], 'bool'],
    'is_digit': [[], 'bool'],
    'is_numeric': [[], 'bool'],
    'is_space': [[], 'bool']
}


def is_snake_case(value):
    return SNAKE_CASE_RE.match(value) is not None


def is_upper_snake_case(value):
    return UPPER_SNAKE_CASE_RE.match(value) is not None


def is_pascal_case(value):
    return PASCAL_CASE_RE.match(value)


def is_primitive_type(mys_type):
    if not isinstance(mys_type, str):
        return False

    return mys_type in PRIMITIVE_TYPES


def is_integer_type(mys_type):
    if not isinstance(mys_type, str):
        return False

    return mys_type in INTEGER_TYPES


def is_float_type(mys_type):
    if not isinstance(mys_type, str):
        return False

    return mys_type in ['f32', 'f64']


def dot2ns(name):
    # Hack...
    if name not in BUILTIN_CALLS and name != 'Error':
        name = f'mys.{name}'

    # ToDo: Super hack...
    if name == 'mys.fiber.lib.Fiber':
        name = 'Fiber'

    return name.replace('.', '::')


def make_name(name):
    return make_cpp_safe_name(name)


def split_dict_mys_type(mys_type):
    key_mys_type = list(mys_type.keys())[0]
    value_mys_type = list(mys_type.values())[0]

    return key_mys_type, value_mys_type


def make_relative_import_absolute(module_levels, module, node):
    prefix = '.'.join(module_levels[0:-node.level])

    if not prefix:
        raise CompileError('relative import is outside package', node)

    if module is None:
        module = prefix
    else:
        module = f'{prefix}.{module}'

    return module


def is_relative_import(node):
    return node.level > 0


def get_import_from_info(node, module_levels):
    module = node.module

    if is_relative_import(node):
        module = make_relative_import_absolute(module_levels, module, node)

    if '.' not in module:
        module += '.lib'

    if len(node.names) != 1:
        raise CompileError(f'only one import is allowed, found {len(node.names)}',
                           node)

    name = node.names[0]

    if name.asname:
        asname = name.asname
    else:
        asname = name.name

    return module, name.name, asname


def make_shared(cpp_type, values):
    return f'mys::make_shared<{cpp_type}>({values})'


def shared_list_type(cpp_type):
    return f'SharedList<{cpp_type}>'


def make_shared_list(cpp_type, value):
    return (f'mys::make_shared<mys::List<{cpp_type}>>('
            f'std::initializer_list<{cpp_type}>{{{value}}})')


def make_shared_set(cpp_type, value):
    return (f'mys::make_shared<Set<{cpp_type}>>('
            f'std::initializer_list<{cpp_type}>{{{value}}})')


def shared_dict_type(key_cpp_type, value_cpp_type):
    return f'SharedDict<{key_cpp_type}, {value_cpp_type}>'


def make_shared_dict(key_cpp_type, value_cpp_type, items):
    return (f'mys::make_shared<Dict<{key_cpp_type}, {value_cpp_type}>>('
            f'std::initializer_list<robin_hood::pair<{key_cpp_type}, '
            f'{value_cpp_type}>>{{{items}}})')


def shared_set_type(cpp_type):
    return f'SharedSet<{cpp_type}>'


def shared_tuple_type(items):
    return f'SharedTuple<{items}>'


def mys_to_cpp_type(mys_type, context):
    if isinstance(mys_type, tuple):
        items = ', '.join([mys_to_cpp_type(item, context) for item in mys_type])

        return shared_tuple_type(items)
    elif isinstance(mys_type, list):
        item = mys_to_cpp_type(mys_type[0], context)

        return shared_list_type(item)
    elif isinstance(mys_type, dict):
        key_mys_type, value_mys_type = split_dict_mys_type(mys_type)
        key = mys_to_cpp_type(key_mys_type, context)
        value = mys_to_cpp_type(value_mys_type, context)

        return shared_dict_type(key, value)
    elif isinstance(mys_type, set):
        item = mys_to_cpp_type(list(mys_type)[0], context)
        return shared_set_type(item)
    else:
        if mys_type == 'string':
            return 'mys::String'
        elif mys_type == 'bool':
            return 'mys::Bool'
        elif mys_type == 'char':
            return 'mys::Char'
        elif mys_type == 'bytes':
            return 'mys::Bytes'
        elif mys_type == 'regex':
            return 'mys::Regex'
        elif mys_type == 'regexmatch':
            return 'mys::RegexMatch'
        elif context.is_class_or_trait_defined(mys_type):
            return f'mys::shared_ptr<{dot2ns(mys_type)}>'
        elif context.is_enum_defined(mys_type):
            return context.get_enum_type(mys_type)
        else:
            return mys_type


def mys_to_cpp_type_param(mys_type, context):
    cpp_type = mys_to_cpp_type(mys_type, context)

    if not is_primitive_type(mys_type):
        if not context.is_enum_defined(mys_type) and mys_type != 'string':
            cpp_type = f'{cpp_type}'

    return cpp_type


def format_return_type(returns, context):
    if returns is not None:
        return mys_to_cpp_type(returns, context)
    else:
        return 'void'


def format_method_name(method, class_name):
    if method.name == '__init__':
        return class_name
    elif method.name == '__del__':
        return f'~{class_name}'
    else:
        return method.name


def format_default(name, param_name, return_cpp_type):
    return f'{return_cpp_type} {name}_{param_name}_default()'


def format_default_call(full_name, param_name):
    return f'{dot2ns(full_name)}_{param_name}_default()'


def is_public(name):
    return not is_private(name)


def is_private(name):
    return name.startswith('_')


def has_docstring(node):
    """Retuns true if given function or method has a docstring.

    """

    return ast.get_docstring(node) is not None


class IntegerLiteralVisitor(ast.NodeVisitor):

    def visit_BinOp(self, node):
        return self.visit(node.left) and self.visit(node.right)

    def visit_UnaryOp(self, node):
        return self.visit(node.operand)

    def visit_Constant(self, node):
        if isinstance(node.value, bool):
            return False
        else:
            return isinstance(node.value, int)

    def generic_visit(self, node):
        return False


def is_integer_literal(node):
    return IntegerLiteralVisitor().visit(node)


class FloatLiteralVisitor(IntegerLiteralVisitor):

    def visit_Constant(self, node):
        return isinstance(node.value, float)


def is_float_literal(node):
    return FloatLiteralVisitor().visit(node)


class MakeIntegerLiteralVisitor(ast.NodeVisitor):

    def __init__(self, type_name):
        self.type_name = type_name
        self.factor = 1

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        op_class = type(node.op)

        return format_binop(left, right, op_class, True)

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.USub):
            factor = -1
        else:
            factor = 1

        self.factor *= factor

        try:
            value = self.visit(node.operand)
        except CompileError as e:
            e.lineno = node.lineno
            e.offset = node.col_offset

            raise e

        self.factor *= factor

        if not isinstance(node.op, (ast.UAdd, ast.USub)):
            op = OPERATORS[type(node.op)]
            value = f'{op}{value}'

        return value

    def visit_Constant(self, node):
        value = node.value * self.factor

        if self.type_name == 'u8':
            if 0 <= value <= 0xff:
                return str(value)
        elif self.type_name == 'u16':
            if 0 <= value <= 0xffff:
                return str(value)
        elif self.type_name == 'u32':
            if 0 <= value <= 0xffffffff:
                return str(value)
        elif self.type_name == 'u64':
            if 0 <= value <= 0xffffffffffffffff:
                return f'u64({value}ull)'
        elif self.type_name == 'i8':
            if -0x80 <= value <= 0x7f:
                return str(value)
        elif self.type_name == 'i16':
            if -0x8000 <= value <= 0x7fff:
                return str(value)
        elif self.type_name == 'i32':
            if -0x80000000 <= value <= 0x7fffffff:
                return str(value)
        elif self.type_name == 'i64':
            if -0x7fffffffffffffff <= value <= 0x7fffffffffffffff:
                return str(value)
            if -0x8000000000000000 == value:
                # g++ warns for -0x8000000000000000.
                return '(-0x7fffffffffffffff - 1)'
        elif self.type_name is None:
            raise CompileError("integers cannot be None", node)
        else:
            mys_type = format_mys_type(self.type_name)

            raise CompileError(f"cannot convert integer to '{mys_type}'", node)

        raise CompileError(
            f"integer literal out of range for '{self.type_name}'",
            node)


def make_integer_literal(type_name, node):
    return MakeIntegerLiteralVisitor(type_name).visit(node)


class MakeFloatLiteralVisitor(MakeIntegerLiteralVisitor):

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        op_class = type(node.op)

        return format_binop(left, right, op_class, False)

    def visit_Constant(self, node):
        value = node.value * self.factor

        if self.type_name == 'f32':
            return str(value)
        elif self.type_name == 'f64':
            return str(value)
        elif self.type_name is None:
            raise CompileError("float cannot be None", node)
        else:
            mys_type = format_mys_type(self.type_name)

            raise CompileError(f"cannot convert float to '{mys_type}'", node)

        raise CompileError(
            f"float literal out of range for '{self.type_name}'",
            node)


def make_float_literal(type_name, node):
    return MakeFloatLiteralVisitor(type_name).visit(node)


def format_binop(left, right, op_class, is_integer=True):
    if op_class == ast.Pow:
        if is_integer:
            return f'ipow({left}, {right})'
        else:
            return f'std::pow({left}, {right})'
    elif op_class in [ast.Mod, ast.Div]:
        op = OPERATORS[op_class]

        return f'({left} {op} mys::denominator_not_zero({right}))'
    else:
        op = OPERATORS[op_class]

        return f'({left} {op} {right})'


def format_mys_type(mys_type):
    if isinstance(mys_type, tuple):
        if len(mys_type) == 1:
            items = f'{format_mys_type(mys_type[0])}, '
        else:
            items = ', '.join([format_mys_type(item) for item in mys_type])

        return f'({items})'
    elif isinstance(mys_type, list):
        item = format_mys_type(mys_type[0])

        return f'[{item}]'
    elif isinstance(mys_type, dict):
        key_mys_type, value_mys_type = split_dict_mys_type(mys_type)
        key = format_mys_type(key_mys_type)
        value = format_mys_type(value_mys_type)

        return f'{{{key}: {value}}}'
    elif isinstance(mys_type, set):
        item = format_mys_type(list(mys_type)[0])
        return f'{{{item}}}'
    elif isinstance(mys_type, GenericType):
        types = ', '.join(format_mys_type(type) for type in mys_type.types)

        return f'{mys_type.name}[{types}]'
    else:
        return str(mys_type)


def raise_types_differs(left_mys_type, right_mys_type, node):
    left = format_mys_type(left_mys_type)
    right = format_mys_type(right_mys_type)

    raise CompileError(f"types '{left}' and '{right}' differs", node)


def raise_if_types_differs(left_mys_type, right_mys_type, node):
    if left_mys_type != right_mys_type:
        raise_types_differs(left_mys_type, right_mys_type, node)


def indent_lines(lines):
    return ['    ' + line for line in lines if line]


def indent(string):
    return '\n'.join(indent_lines(string.splitlines()))


def dedent(string):
    return '\n'.join([line[4:] for line in string.splitlines() if line])


def make_types_string_parts(mys_types):
    """Makes given list of mys types a new list with separators inserted
    for unique function names.

    """

    parts = []

    for mys_type in mys_types:
        if isinstance(mys_type, str):
            parts.append(mys_type.replace('.', '_'))
        elif isinstance(mys_type, tuple):
            parts.append('tb')
            parts += make_types_string_parts(mys_type)
            parts.append('te')
        elif isinstance(mys_type, list):
            parts.append('lb')
            parts += make_types_string_parts(mys_type)
            parts.append('le')
        elif isinstance(mys_type, dict):
            key_mys_type, value_mys_type = split_dict_mys_type(mys_type)
            parts.append('db')
            parts += make_types_string_parts([key_mys_type])
            parts.append('cn')
            parts += make_types_string_parts([value_mys_type])
            parts.append('de')
        else:
            raise Exception(str(mys_type))

    return parts


def make_function_name(name, param_types, returns):
    """Create a function name from given name, parameters and return type.

    """

    parts = [name]
    parts += make_types_string_parts(param_types)

    if returns is not None:
        parts.append('r')
        parts += make_types_string_parts([returns])

    return '_'.join(parts)


def raise_wrong_types(actual_mys_type, expected_mys_type, node):
    if is_primitive_type(expected_mys_type) and actual_mys_type is None:
        raise CompileError(f"'{expected_mys_type}' cannot be None", node)
    else:
        actual = format_mys_type(actual_mys_type)
        expected = format_mys_type(expected_mys_type)

        raise CompileError(f"expected a '{expected}', got a '{actual}'", node)


def raise_if_wrong_types(actual_mys_type, expected_mys_type, node, context):
    if actual_mys_type == expected_mys_type:
        return

    if actual_mys_type is None:
        if context.is_class_or_trait_defined(expected_mys_type):
            return
        elif expected_mys_type in ['string', 'bytes']:
            return
        elif isinstance(expected_mys_type, (list, tuple, dict)):
            return

    raise_wrong_types(actual_mys_type, expected_mys_type, node)


def raise_if_wrong_visited_type(context, expected_mys_type, node):
    raise_if_wrong_types(context.mys_type,
                         expected_mys_type,
                         node,
                         context)



import copy

from ..parser import ast
from .definitions import Class
from .definitions import Function
from .utils import CompileError
from .utils import GenericType
from .utils import make_name
from .utils import make_types_string_parts
from .utils import mys_to_cpp_type_param
from .utils import split_dict_mys_type


def replace_generic_types(generic_types, mys_type, chosen_types):
    if isinstance(mys_type, str):
        for generic_type, chosen_type in zip(generic_types, chosen_types):
            if mys_type == generic_type:
                return chosen_type

    return mys_type


class SpecializeGenericType:

    def __init__(self, generic_type, chosen_type, node=None):
        self.generic_type = generic_type
        self.chosen_type = chosen_type
        self.node = node

    def replace(self, mys_type):
        """Replaces all occurrences of generic types with chosen types.

        """

        if isinstance(mys_type, str):
            if mys_type == self.generic_type:
                mys_type = self.chosen_type
        elif isinstance(mys_type, dict):
            key_mys_type, value_mys_type = split_dict_mys_type(mys_type)
            mys_type = {self.replace(key_mys_type): self.replace(value_mys_type)}
        elif isinstance(mys_type, set):
            mys_type = {self.replace(list(mys_type)[0])}
        elif isinstance(mys_type, list):
            mys_type = [self.replace(mys_type[0])]
        elif isinstance(mys_type, tuple):
            mys_type = tuple(self.replace(item_mys_type)
                             for item_mys_type in mys_type)
        else:
            raise Exception('generic type not supported')

        return mys_type


class SpecializeTypeTransformer(ast.NodeTransformer):
    """Traverses given generic node and replaces given generic types with
    given chosen types.

    """

    def __init__(self, generic_types, chosen_types):
        self.generic_to_specialized_type = dict(zip(generic_types, chosen_types))

    def visit_Name(self, node):
        node.id = self.generic_to_specialized_type.get(node.id, node.id)

        return node


def specialize_function(function, specialized_full_name, chosen_types, node):
    """Returns a copy of the function object with all generic types
    replaced with chosen types.

    """

    returns = function.returns
    args = copy.deepcopy(function.args)
    actual_ntypes = len(chosen_types)
    expected_ntypes = len(function.generic_types)

    if actual_ntypes != expected_ntypes:
        raise CompileError(
            f'expected {expected_ntypes} type, got {actual_ntypes}',
            node.func.slice)

    for generic_type, chosen_type in zip(function.generic_types, chosen_types):
        if returns is not None:
            returns = SpecializeGenericType(generic_type,
                                            chosen_type,
                                            function.node).replace(returns)

        for param, _ in args:
            param.type = SpecializeGenericType(generic_type,
                                               chosen_type,
                                               function.node).replace(param.type)

    node = copy.deepcopy(function.node)
    node.name = specialized_full_name
    node = SpecializeTypeTransformer(function.generic_types,
                                     chosen_types).visit(node)

    return Function(specialized_full_name,
                    [],
                    function.raises,
                    function.is_test,
                    args,
                    returns,
                    node,
                    function.module_name)


def specialize_class(definitions, specialized_name, chosen_types, node):
    """Returns a copy of the class object with all generic types replaced
    with chosen types.

    """

    members = copy.deepcopy(definitions.members)
    methods = copy.deepcopy(definitions.methods)
    actual_ntypes = len(chosen_types)
    expected_ntypes = len(definitions.generic_types)

    if actual_ntypes != expected_ntypes:
        raise CompileError(
            f'expected {expected_ntypes} type, got {actual_ntypes}',
            node.slice)

    for generic_type, chosen_type in zip(definitions.generic_types, chosen_types):
        for member in members.values():
            member.type = SpecializeGenericType(generic_type,
                                                chosen_type).replace(member.type)

        for class_methods in methods.values():
            for method in class_methods:
                if method.returns is not None:
                    method.returns = SpecializeGenericType(
                        generic_type,
                        chosen_type).replace(method.returns)

                for param, _node in method.args:
                    param.type = SpecializeGenericType(
                        generic_type,
                        chosen_type).replace(param.type)

                method.node = SpecializeTypeTransformer(
                    definitions.generic_types,
                    chosen_types).visit(method.node)

    return Class(specialized_name,
                 [],
                 members,
                 methods,
                 definitions.functions,
                 definitions.implements,
                 definitions.node,
                 definitions.module_name)


def add_generic_class(node, context):
    name = node.value.id
    full_name = context.make_full_name(name)
    chosen_types = find_chosen_types(node, context)
    specialized_name, specialized_full_name = make_generic_name(
        name,
        full_name,
        chosen_types)
    definitions = context.get_class_definitions(full_name)

    if context.is_specialized_class_defined(specialized_full_name):
        specialized_class = context.get_specialized_class(
            specialized_full_name)
    else:
        specialized_class = specialize_class(definitions,
                                             specialized_name,
                                             chosen_types,
                                             node)
        context.define_specialized_class(specialized_full_name,
                                         specialized_class,
                                         node)

    context.define_class(specialized_name,
                         specialized_full_name,
                         specialized_class)

    return specialized_class, specialized_full_name


def make_generic_name(name, full_name, chosen_types):
    joined_chosen_types = '_'.join(make_types_string_parts(chosen_types))
    specialized_name = f'{name}_{joined_chosen_types}'
    specialized_full_name = f'{full_name}_{joined_chosen_types}'

    return specialized_name, specialized_full_name


def find_chosen_types(node, context):
    types_slice = node.slice
    chosen_types = [
        TypeVisitor(context).visit(type_node)
        for type_node in fix_chosen_types(types_slice, context.source_lines)
    ]

    return chosen_types


def format_parameters(args, context):
    parameters = []

    for param, _ in args:
        if isinstance(param.type, GenericType):
            param_type = add_generic_class(param.type.node, context)[1]
        else:
            param_type = param.type

        cpp_type = mys_to_cpp_type_param(param_type, context)
        parameters.append(f'{cpp_type} {make_name(param.name)}')

    if parameters:
        return ', '.join(parameters)
    else:
        return 'void'


def fix_chosen_types(slice_node, source_lines):
    """Returns a list of nodes that represents the chosen types.

    """

    if isinstance(slice_node, ast.Tuple):
        first = slice_node.elts[0]
        second = slice_node.elts[1]

        if slice_node.lineno != second.lineno:
            raise Exception('internal error')

        source = source_lines[slice_node.lineno - 1]
        opening = source[slice_node.col_offset:first.col_offset]
        closing = source[first.end_col_offset:second.col_offset]

        if opening.count('(') > closing.count(')'):
            types = [slice_node]
        else:
            types = slice_node.elts
    else:
        types = [slice_node]

    return types


class TypeVisitor(ast.NodeVisitor):

    def __init__(self, context):
        self.context = context

    def visit_Name(self, node):
        name = node.id

        if self.context.is_class_defined(name):
            name = self.context.make_full_name(name)
        elif self.context.is_enum_defined(name):
            name = self.context.make_full_name(name)
        elif self.context.is_trait_defined(name):
            name = self.context.make_full_name(name)

        return name

    def visit_List(self, node):
        nitems = len(node.elts)

        if nitems != 1:
            raise CompileError(f"expected 1 type in list, got {nitems}", node)

        return [self.visit(elem) for elem in node.elts]

    def visit_Tuple(self, node):
        return tuple(self.visit(elem) for elem in node.elts)

    def visit_Dict(self, node):
        return {node.keys[0].id: self.visit(node.values[0])}

    def visit_Set(self, node):
        nitems = len(node.elts)

        if nitems != 1:
            raise CompileError(f"expected 1 type in set, got {nitems}", node)

        return {self.visit(node.elts[0])}

    def visit_Subscript(self, node):
        return add_generic_class(node, self.context)[1]



import glob
import os
import shutil
from unittest.mock import patch

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import remove_ansi
from .utils import remove_build_directory
from .utils import transpile_source


class Test(TestCase):

    def test_style(self):
        name = 'test_style'
        remove_build_directory(name)

        shutil.copytree('tests/files/style', f'tests/build/{name}')

        with Path(f'tests/build/{name}'):
            mod_mtime = os.stat('src/mod.mys').st_mtime

            # Ensure that there are trailing whitespaces in the file.
            with open('src/trailing_whitespaces.mys', 'r') as fin:
                source = fin.read()

            self.assertEqual(source.count(' \n'), len(source.split('\n')) - 1)

            with patch('sys.argv', ['mys', '-d', 'style']):
                mys.cli.main()

            # Files that are already styled should not be written to.
            self.assertEqual(os.stat('src/mod.mys').st_mtime, mod_mtime)

            for path in glob.glob('src/**/*.mys', recursive=True):
                self.assert_files_equal(path, f'../../files/style/styled-{path}')

    def test_tabs_not_allowed_as_indentation(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def foo():\n'
                             "\tpass\n")

        self.assertEqual(
            remove_ansi(str(cm.exception).replace('\t', '')),
            '  File "<string>", line 2\n'
            '    pass\n'
            '    ^\n'
            'SyntaxError: indentations must be spaces, not tabs\n')

    def test_indent_2_spaces(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def foo():\n'
                             "  pass\n")

        self.assertEqual(
            remove_ansi(str(cm.exception).replace('\t', '')),
            '  File "<string>", line 2\n'
            '    pass\n'
            '    ^\n'
            'SyntaxError: indentations must be 4 spaces\n')

    def test_indent_2_spaces_in_if(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def foo():\n'
                             "    if True:\n"
                             "      pass\n")

        self.assertEqual(
            remove_ansi(str(cm.exception).replace('\t', '')),
            '  File "<string>", line 3\n'
            '    pass\n'
            '    ^\n'
            'SyntaxError: indentations must be 4 spaces\n')



from ..parser import ast


class ConstantVisitor(ast.NodeVisitor):

    def __init__(self):
        self.is_constant = True

    def visit_Constant(self, node):
        pass

    def visit_UnaryOp(self, node):
        self.visit(node.operand)

    def visit_List(self, node):
        for item in node.elts:
            self.visit(item)

    def visit_Tuple(self, node):
        for item in node.elts:
            self.visit(item)

    def visit_Dict(self, node):
        for item in node.keys:
            self.visit(item)

        for item in node.values:
            self.visit(item)

    def generic_visit(self, node):
        self.is_constant = False


def is_constant(node):
    """Returns true if given node is a "constant", that is, is does not
    contain variables, enums, traits or classes.

    """

    visitor = ConstantVisitor()
    visitor.visit(node)

    return visitor.is_constant



from .utils import TestCase


class Test(TestCase):

    def test_while_non_bool(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    while 1:\n'
            '        pass\n',
            '  File "", line 2\n'
            "        while 1:\n"
            '              ^\n'
            "CompileError: expected a 'bool', got a 'i64'\n")

    def test_bare_integer_in_while(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    while True:\n'
            '        1\n',
            '  File "", line 3\n'
            '            1\n'
            '            ^\n'
            "CompileError: bare integer\n")

    def test_variable_defined_in_while_can_not_be_used_after_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    while False:\n'
            '        v = 1\n'
            '    print(v)\n',
            '  File "", line 4\n'
            '        print(v)\n'
            '              ^\n'
            "CompileError: undefined variable 'v'\n")



from ..parser import ast
from .definitions import GenericType
from .generics import TypeVisitor
from .generics import add_generic_class
from .generics import fix_chosen_types
from .generics import replace_generic_types
from .utils import BUILTIN_CALLS
from .utils import BUILTIN_ERRORS
from .utils import BYTES_METHODS
from .utils import LIST_METHODS
from .utils import NUMBER_TYPES
from .utils import OPERATORS_TO_METHOD
from .utils import REGEX_METHODS
from .utils import REGEXMATCH_METHODS
from .utils import SET_METHODS
from .utils import STRING_METHODS
from .utils import CompileError
from .utils import InternalError
from .utils import is_primitive_type
from .utils import is_snake_case
from .utils import make_integer_literal
from .utils import raise_if_types_differs
from .utils import split_dict_mys_type


def mys_to_value_type(mys_type):
    if isinstance(mys_type, tuple):
        return tuple(mys_to_value_type(item) for item in mys_type)
    elif isinstance(mys_type, list):
        return [mys_to_value_type(item) for item in mys_type]
    elif isinstance(mys_type, dict):
        key_mys_type, value_mys_type = split_dict_mys_type(mys_type)

        return Dict(mys_to_value_type(key_mys_type),
                    mys_to_value_type(value_mys_type))
    elif isinstance(mys_type, set):
        return Set(mys_to_value_type(list(mys_type)[0]))
    else:
        return mys_type


def format_value_type(value_type):
    if isinstance(value_type, tuple):
        if len(value_type) == 1:
            items = f'{format_value_type(value_type[0])}, '
        else:
            items = ', '.join([format_value_type(item) for item in value_type])

        return f'({items})'
    elif isinstance(value_type, list):
        if len(value_type) == 1:
            return f'[{format_value_type(value_type[0])}]'
        else:
            return '/'.join([format_value_type(item) for item in value_type])
    elif isinstance(value_type, dict):
        raise Exception('not implemented')
    else:
        return value_type


def intersection_of(type_1, type_2, node):
    """Find the intersection of given visited types.

    """

    if type_1 is None:
        if is_primitive_type(type_2):
            raise CompileError(f"'{type_2}' cannot be None", node)
        else:
            return type_2, type_2
    elif type_2 is None:
        if is_primitive_type(type_1):
            raise CompileError(f"'{type_1}' cannot be None", node)
        else:
            return type_1, type_1
    elif type_1 == type_2:
        return type_1, type_2
    elif isinstance(type_1, str) and isinstance(type_2, str):
        raise_if_types_differs(type_1, type_2, node)
    elif isinstance(type_1, tuple) and isinstance(type_2, tuple):
        if len(type_1) != len(type_2):
            return None, None
        else:
            new_type_1 = []
            new_type_2 = []

            for item_type_1, item_type_2 in zip(type_1, type_2):
                item_type_1, item_type_2 = intersection_of(item_type_1,
                                                           item_type_2,
                                                           node)

                if item_type_1 is None or item_type_2 is None:
                    return None, None

                new_type_1.append(item_type_1)
                new_type_2.append(item_type_2)

            return tuple(new_type_1), tuple(new_type_2)
    elif isinstance(type_1, Set) and isinstance(type_2, Set):
        value_type_1, value_type_2 = intersection_of(type_1.value_type,
                                                     type_2.value_type,
                                                     node)
        return Set(value_type_1), Set(value_type_2)
    elif isinstance(type_1, Set) and isinstance(type_2, Dict):
        if type_2.key_type is not None or type_2.value_type is not None:
            raise CompileError(f"cannot convert '{type_1}' to '{type_2}'", node)
        return type_1, type_1
    elif isinstance(type_1, Dict) and isinstance(type_2, Dict):
        if type_1.key_type is None and type_2.key_type is not None:
            type_1.key_type = type_2.key_type
        elif type_1.key_type is not None and type_2.key_type is None:
            type_2.key_type = type_1.key_type

        key_value_type_1, key_value_type_2 = intersection_of(type_1.key_type,
                                                             type_2.key_type,
                                                             node)

        if type_1.value_type is None and type_2.value_type is not None:
            type_1.value_type = type_2.value_type
        elif type_1.value_type is not None and type_2.value_type is None:
            type_2.value_type = type_1.value_type

        value_value_type_1, value_value_type_2 = intersection_of(type_1.value_type,
                                                                 type_2.value_type,
                                                                 node)

        return (Dict(key_value_type_1, value_value_type_1),
                Dict(key_value_type_2, value_value_type_2))
    elif isinstance(type_1, str):
        if not isinstance(type_2, list):
            return None, None
        elif type_1 not in type_2:
            type_1 = format_value_type(type_1)
            type_2 = format_value_type(type_2)

            raise CompileError(f"cannot convert '{type_1}' to '{type_2}'", node)
        else:
            return type_1, type_1
    elif isinstance(type_2, str):
        if not isinstance(type_1, list):
            return None, None
        elif type_2 not in type_1:
            type_1 = format_value_type(type_1)
            type_2 = format_value_type(type_2)

            raise CompileError(f"cannot convert '{type_1}' to '{type_2}'", node)
        else:
            return type_2, type_2
    elif isinstance(type_1, list) and isinstance(type_2, list):
        if len(type_1) == 0 and len(type_2) == 0:
            return [], []
        elif len(type_1) == 1 and len(type_2) == 1:
            item_type_1, item_type_2 = intersection_of(type_1[0], type_2[0], node)

            return [item_type_1], [item_type_2]
        elif len(type_1) == 1 and len(type_2) == 0:
            return type_1, type_1
        elif len(type_2) == 1 and len(type_1) == 0:
            return type_2, type_2
        elif len(type_1) == 1 and len(type_2) > 1:
            type_1 = format_value_type(type_1)
            type_2 = format_value_type(type_2)

            raise CompileError(f"cannot convert '{type_1}' to '{type_2}'", node)
        elif len(type_2) == 1 and len(type_1) > 1:
            type_1 = format_value_type(type_1)
            type_2 = format_value_type(type_2)

            raise CompileError(f"cannot convert '{type_1}' to '{type_2}'", node)
        else:
            new_type_1 = []
            new_type_2 = []

            for item_type_1 in type_1:
                for item_type_2 in type_2:
                    if isinstance(item_type_1, str) and isinstance(item_type_2, str):
                        if item_type_1 != item_type_2:
                            continue
                    else:
                        item_type_1, item_type_2 = intersection_of(item_type_1,
                                                                   item_type_2,
                                                                   node)

                        if item_type_1 is None or item_type_2 is None:
                            continue

                    new_type_1.append(item_type_1)
                    new_type_2.append(item_type_2)

            if len(new_type_1) == 0:
                type_1 = format_value_type(type_1)
                type_2 = format_value_type(type_2)

                raise CompileError(f"cannot convert '{type_1}' to '{type_2}'", node)
            elif len(new_type_1) == 1:
                return new_type_1[0], new_type_2[0]
            else:
                return new_type_1, new_type_2
    else:
        raise InternalError(f"specialize types {type_1}, {type_2}", node)


def reduce_type(value_type):
    if isinstance(value_type, list):
        if len(value_type) == 0:
            return ['bool']
        elif len(value_type) == 1:
            return [reduce_type(value_type[0])]
        else:
            return reduce_type(value_type[0])
    elif isinstance(value_type, tuple):
        values = []

        for item in value_type:
            values.append(reduce_type(item))

        return tuple(values)
    elif isinstance(value_type, str):
        return value_type
    elif isinstance(value_type, Dict):
        return {reduce_type(value_type.key_type): reduce_type(value_type.value_type)}
    elif isinstance(value_type, Set):
        return {reduce_type(value_type.value_type)}
    elif value_type is None:
        return None
    else:
        raise Exception("Bad reduce")


class Set:
    def __init__(self, value_type):
        self.value_type = value_type

    def __str__(self):
        return f'Set({self.value_type})'


class Dict:

    def __init__(self, key_type, value_type):
        self.key_type = key_type
        self.value_type = value_type

    def __str__(self):
        return f'Dict({self.key_type}, {self.value_type})'


class ValueTypeVisitor(ast.NodeVisitor):
    """Find the type of given value.

    """

    def __init__(self, context):
        self.context = context
        self.factor = 1

    def visit_BoolOp(self, _node):
        return 'bool'

    def visit_JoinedStr(self, _node):
        return 'string'

    def visit_Compare(self, _node):
        return 'bool'

    def visit_BinOp(self, node):
        left_value_type = self.visit(node.left)
        right_value_type = self.visit(node.right)

        if self.context.is_class_defined(left_value_type):
            method = self.find_operator_method(left_value_type,
                                               OPERATORS_TO_METHOD[type(node.op)],
                                               node)

            return method.returns

        if right_value_type == 'string':
            return 'string'
        elif left_value_type == 'string':
            return 'string'
        else:
            return intersection_of(left_value_type, right_value_type, node)[0]

    def visit_Slice(self, node):
        lower = None
        upper = None
        step = None

        if node.lower:
            lower = self.visit(node.lower)

        if node.upper:
            upper = self.visit(node.upper)

        if node.step:
            step = self.visit(node.step)

        return lower, upper, step

    def visit_Subscript(self, node):
        value_type = mys_to_value_type(self.visit(node.value))

        if isinstance(value_type, list):
            value_type = value_type[0]
        elif isinstance(value_type, tuple):
            index = make_integer_literal('i64', node.slice)
            value_type = value_type[int(index)]
        elif isinstance(value_type, Dict):
            value_type = value_type.value_type
        elif value_type == 'string':
            slice_type = self.visit(node.slice)
            if isinstance(slice_type, tuple):
                value_type = 'string'
            else:
                value_type = 'char'
        elif value_type == 'bytes':
            value_type = 'u8'
        else:
            raise Exception('todo')

        return value_type

    def visit_IfExp(self, node):
        return self.visit(node.body)

    def visit_Attribute(self, node):
        name = node.attr

        if isinstance(node.value, ast.Name):
            value = node.value.id

            if self.context.is_enum_defined(value):
                value_type = self.context.make_full_name(value)
            elif self.context.is_local_variable_defined(value):
                value_type = self.context.get_local_variable_type(value)
            elif self.context.is_global_variable_defined(value):
                value_type = self.context.get_global_variable_type(value)
            else:
                raise InternalError("attribute", node)
        else:
            value_type = self.visit(node.value)

        if self.context.is_class_defined(value_type):
            definitions = self.context.get_class_definitions(value_type)

            if name in definitions.members:
                value_type = definitions.members[name].type
            else:
                raise CompileError(
                    f"class '{value_type}' has no member '{name}'",
                    node)

        if isinstance(value_type, dict):
            value_type = Dict(list(value_type.keys())[0],
                              list(value_type.values())[0])
        elif isinstance(value_type, set):
            value_type = Set(list(value_type)[0])

        return value_type

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.USub):
            factor = -1
        else:
            factor = 1

        self.factor *= factor
        value = self.visit(node.operand)
        self.factor *= factor

        return value

    def visit_Constant(self, node):
        if isinstance(node.value, bool):
            return 'bool'
        elif isinstance(node.value, int):
            types = ['i64', 'i32', 'i16', 'i8']

            if self.factor == 1:
                types += ['u64', 'u32', 'u16', 'u8']

            return types
        elif isinstance(node.value, float):
            return ['f64', 'f32']
        elif isinstance(node.value, str):
            return 'string'
        elif isinstance(node.value, bytes):
            return 'bytes'
        elif isinstance(node.value, tuple):
            if len(node.value) == 2:
                return 'regex'
            elif len(node.value) == 3:
                return 'char'
            else:
                raise Exception('todo')
        elif isinstance(node.value, complex):
            raise CompileError('complex numbers are not supported', node)
        elif node.value is None:
            return None
        else:
            raise Exception('todo')

    def visit_Name(self, node):
        name = node.id

        if name == '__unique_id__':
            return 'i64'
        elif name == '__line__':
            return 'u64'
        elif name == '__name__':
            return 'string'
        elif name == '__file__':
            return 'string'
        elif self.context.is_local_variable_defined(name):
            value_type = self.context.get_local_variable_type(name)

            return mys_to_value_type(value_type)
        elif self.context.is_global_variable_defined(name):
            value_type = self.context.get_global_variable_type(name)

            return mys_to_value_type(value_type)
        else:
            raise CompileError(f"undefined variable '{name}'", node)

    def visit_List(self, node):
        if len(node.elts) == 0:
            return []

        item_type = self.visit(node.elts[0])

        for item in node.elts[1:]:
            item_type, _ = intersection_of(item_type, self.visit(item), item)

        return [item_type]

    def visit_Tuple(self, node):
        return tuple(self.visit(elem) for elem in node.elts)

    def visit_Dict(self, node):
        if len(node.keys) > 0:
            return Dict(self.visit(node.keys[0]), self.visit(node.values[0]))
        else:
            return Dict(None, None)

    def visit_Set(self, node):
        if len(node.elts) == 0:
            return Set([])

        item_type = self.visit(node.elts[0])

        for item in node.elts[1:]:
            item_type, _ = intersection_of(item_type, self.visit(item), item)

        return Set(item_type)

    def visit_call_params_keywords(self, function, node):
        keyword_args = {}
        params = {param.name for param, _ in function.args}
        positional_params_names = [
            param.name for param, _ in function.args[:len(node.args)]
        ]

        if node.keywords:
            for keyword in node.keywords:
                param_name = keyword.arg

                if param_name not in params:
                    return None

                if param_name in positional_params_names:
                    return None

                if param_name in keyword_args:
                    return None

                keyword_args[param_name] = keyword.value

        return keyword_args

    def visit_call_params(self, function, node):
        """Returns true if given function can be called with given parameters,
        false otherwise.

        """

        keyword_args = self.visit_call_params_keywords(function, node)

        if keyword_args is None:
            return False

        for i, (param, default) in enumerate(function.args):
            if i < len(node.args):
                try:
                    intersection_of(self.visit(node.args[i]), param.type, node)
                except CompileError:
                    return False
            else:
                value = keyword_args.get(param.name)

                if value is None:
                    if default is None:
                        return False
                else:
                    try:
                        intersection_of(self.visit(value), param.type, node)
                    except CompileError:
                        return False

        min_args = len([default for _, default in function.args if default is None])
        nargs = len(node.args) + len(node.keywords)

        return min_args <= nargs <= len(function.args)

    def find_called_function(self, name, node):
        """Find called function. Raises errors if not found or if multiple
        functions matches.

        """

        functions = self.context.get_functions(name)

        if len(functions) == 1:
            return functions[0]
        else:
            matching_functions = []

            for function in functions:
                if self.visit_call_params(function, node):
                    matching_functions.append(function)

            if len(matching_functions) != 1:
                raise CompileError('ambigious function call', node)

            return matching_functions[0]

    def find_called_method(self, class_name, name, node):
        """Find called method. Raises errors if not found or if multiple
        methods matches.

        """

        definitions = self.context.get_class_definitions(class_name)
        methods = definitions.methods[name]

        if len(methods) == 1:
            return methods[0]
        else:
            matching_methods = []

            for method in methods:
                if self.visit_call_params(method, node):
                    matching_methods.append(method)

            if len(matching_methods) > 1:
                raise CompileError('ambigious method call', node)
            elif len(matching_methods) == 0:
                raise CompileError(
                    f"class '{class_name}' has no method '{name}'",
                    node)

            return matching_methods[0]

    def find_operator_method(self, left_value_type, method, node):
        return self.find_called_method(
            left_value_type,
            method,
            ast.Call(func=ast.Name(id=method),
                     args=[node.right],
                     keywords=[],
                     lineno=node.left.lineno,
                     col_offset=node.left.col_offset))

    def find_aug_operator_method(self, target_mys_type, method, node):
        return self.find_called_method(
            target_mys_type,
            method,
            ast.Call(func=ast.Name(id=method),
                     args=[node.value],
                     keywords=[],
                     lineno=node.target.lineno,
                     col_offset=node.target.col_offset))

    def visit_call_function(self, name, node):
        function = self.find_called_function(name, node)

        return mys_to_value_type(function.returns)

    def visit_call_class(self, mys_type, _node):
        return mys_type

    def visit_call_enum(self, mys_type, _node):
        return mys_type

    def visit_call_builtin(self, name, node):
        if name in NUMBER_TYPES:
            return name
        elif name == 'len':
            return 'u64'
        elif name in ['str', 'string']:
            return 'string'
        elif name == 'char':
            return 'char'
        elif name == 'regex':
            return 'regex'
        elif name == 'set':
            value_type = self.visit(node.args[0])
            if isinstance(value_type, list):
                return Set(value_type[0])
            else:
                raise InternalError(f"set('{value_type}') not supported", node)
        elif name == 'list':
            value_type = self.visit(node.args[0])

            if isinstance(value_type, Dict):
                return [(value_type.key_type, value_type.value_type)]
            else:
                raise InternalError(f"list('{value_type}') not supported", node)
        elif name in ['min', 'max', 'sum']:
            value_type = self.visit(node.args[0])

            if isinstance(value_type, list):
                return value_type[0]
            elif isinstance(value_type, Set):
                return value_type.value_type
            else:
                return value_type
        elif name == 'abs':
            return self.visit(node.args[0])
        elif name == 'range':
            return ['i64']
        elif name == 'enumerate':
            # ???
            return [('i64', self.visit(node.args[0]))]
        elif name == 'zip':
            # ???
            return [self.visit(node.args[0])]
        elif name == 'slice':
            # ???
            return [self.visit(node.args[0])]
        elif name == 'reversed':
            # ???
            return [self.visit(node.args[0])]
        elif name == 'input':
            return 'string'
        elif name in BUILTIN_ERRORS:
            return name
        elif name == 'bytes':
            return 'bytes'
        elif name == 'default':
            type_name = node.args[0].id

            if self.context.is_class_or_trait_defined(type_name):
                return self.context.make_full_name(type_name)
            else:
                return type_name
        else:
            raise InternalError(f"builtin '{name}' not supported", node)

    def visit_call_method_list(self, name, value_type, node):
        spec = LIST_METHODS.get(name, None)

        if spec is None:
            raise InternalError(f"string method '{name}' not supported", node)

        if spec[1] == '<listtype>':
            return value_type[0]
        else:
            return spec[1]

    def visit_call_method_dict(self, name, value_type, node):
        if name in ['get', 'pop']:
            return value_type.value_type
        elif name == 'keys':
            return [value_type.key_type]
        elif name == 'values':
            return [value_type.value_type]
        else:
            raise InternalError(f"dict method '{name}' not supported", node)

    def visit_call_method_string(self, name, node):
        spec = STRING_METHODS.get(name, None)

        if spec is None:
            raise InternalError(f"string method '{name}' not supported", node)

        return spec[1]

    def visit_call_method_bytes(self, name, node):
        spec = BYTES_METHODS.get(name, None)

        if spec is None:
            raise InternalError(f"bytes method '{name}' not supported", node)

        return spec[1]

    def visit_call_method_set(self, name, value_type, node):
        spec = SET_METHODS.get(name, None)

        if spec is None:
            raise InternalError(f"set method '{name}' not supported", node)

        if spec[1] == 'set':
            return value_type
        else:
            return spec[1]

    def visit_call_method_regexmatch(self, name, node):
        spec = REGEXMATCH_METHODS.get(name, None)

        if spec is None:
            raise InternalError(f"regexmatch method '{name}' not supported", node)

        return spec[1]

    def visit_call_method_regex(self, name, node):
        spec = REGEX_METHODS.get(name, None)

        if spec is None:
            raise InternalError(f"regex method '{name}' not supported", node)

        return spec[1]

    def visit_call_method_class(self, name, value_type, node):
        method = self.find_called_method(value_type, name, node)
        returns = method.returns

        if isinstance(returns, dict):
            returns = Dict(list(returns.keys())[0], list(returns.values())[0])

        return returns

    def visit_call_method_trait(self, name, value_type, _node):
        method = self.context.get_trait_definitions(value_type).methods[name][0]
        returns = method.returns

        if isinstance(returns, dict):
            returns = Dict(list(returns.keys())[0], list(returns.values())[0])

        return returns

    def visit_call_method_generic(self, name, value_type, node):
        if self.context.is_class_defined(value_type.name):
            value_type = add_generic_class(value_type.node, self.context)[1]

            return self.visit_call_method_class(name, value_type, node)
        else:
            raise InternalError("generic trait not implemented", node)

    def visit_call_method(self, node):
        name = node.func.attr
        value_type = self.visit(node.func.value)

        if isinstance(value_type, list):
            return self.visit_call_method_list(name, value_type, node.func)
        elif isinstance(value_type, Set):
            return self.visit_call_method_set(name, value_type, node.func)
        elif isinstance(value_type, Dict):
            return self.visit_call_method_dict(name, value_type, node.func)
        elif value_type == 'string':
            return self.visit_call_method_string(name, node.func)
        elif value_type == 'regexmatch':
            return self.visit_call_method_regexmatch(name, node.func)
        elif value_type == 'regex':
            return self.visit_call_method_regex(name, node.func)
        elif value_type == 'bytes':
            return self.visit_call_method_bytes(name, node.func)
        elif self.context.is_class_defined(value_type):
            return self.visit_call_method_class(name, value_type, node)
        elif self.context.is_trait_defined(value_type):
            return self.visit_call_method_trait(name, value_type, node)
        elif isinstance(value_type, GenericType):
            return self.visit_call_method_generic(name, value_type, node)
        else:
            raise CompileError("None has no methods", node.func)

    def visit_call_generic_function(self, node):
        name = node.func.value.id
        full_name = self.context.make_full_name(name)
        function = self.context.get_functions(full_name)[0]
        types_slice = node.func.slice
        chosen_types = [
            mys_to_value_type(TypeVisitor(self.context).visit(type_node))
            for type_node in fix_chosen_types(types_slice, self.context.source_lines)
        ]

        return replace_generic_types(function.generic_types,
                                     function.returns,
                                     chosen_types)

    def visit_call_generic_class(self, node):
        return add_generic_class(node.func, self.context)[1]

    def visit_call_generic(self, node):
        if isinstance(node.func.value, ast.Name):
            if self.context.is_class_defined(node.func.value.id):
                return self.visit_call_generic_class(node)
            else:
                return self.visit_call_generic_function(node)
        else:
            raise CompileError("only generic functions and classes are supported",
                               node)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name):
            name = node.func.id

            if name in BUILTIN_CALLS:
                return self.visit_call_builtin(name, node)
            else:
                full_name = self.context.make_full_name(name)

                if full_name is None:
                    if is_snake_case(name):
                        raise CompileError(f"undefined function '{name}'", node)
                    else:
                        raise CompileError(f"undefined class/trait/enum '{name}'",
                                           node)
                elif self.context.is_function_defined(full_name):
                    return self.visit_call_function(full_name, node)
                elif self.context.is_class_defined(full_name):
                    return self.visit_call_class(full_name, node)
                elif self.context.is_enum_defined(full_name):
                    return self.visit_call_enum(full_name, node)
        elif isinstance(node.func, ast.Attribute):
            return self.visit_call_method(node)
        elif isinstance(node.func, ast.Lambda):
            raise CompileError('lambda functions are not supported', node.func)
        elif isinstance(node.func, ast.Subscript):
            return self.visit_call_generic(node)
        else:
            raise CompileError("not callable", node.func)

    def visit_define_local_variables(self, node, mys_type):
        if isinstance(node, ast.Name):
            if not node.id.startswith('_'):
                self.context.define_local_variable(node.id, mys_type, node)
        elif isinstance(node, ast.Tuple):
            for item, item_mys_type in zip(node.elts, mys_type):
                self.visit_define_local_variables(item, item_mys_type)
        else:
            raise CompileError("unsupported type", node)

    def visit_ListComp(self, node):
        if len(node.generators) != 1:
            raise CompileError("only one for-loop allowed", node)

        generator = node.generators[0]
        iter_type = self.visit(generator.iter)

        if isinstance(iter_type, list):
            item_type = iter_type[0]
        elif isinstance(iter_type, Dict):
            item_type = (iter_type.key_type, iter_type.value_type)
        else:
            raise CompileError("unsupported type", node)

        self.context.push()
        self.visit_define_local_variables(generator.target, item_type)
        result_type = [self.visit(node.elt)]
        self.context.pop()

        return result_type

    def visit_DictComp(self, node):
        if len(node.generators) != 1:
            raise CompileError("only one for-loop allowed", node)

        generator = node.generators[0]
        iter_type = self.visit(generator.iter)

        if isinstance(iter_type, list):
            item_type = iter_type[0]
        elif isinstance(iter_type, Dict):
            item_type = (iter_type.key_type, iter_type.value_type)
        else:
            raise CompileError("unsupported type", node)

        self.context.push()
        self.visit_define_local_variables(generator.target, item_type)
        result_type = Dict(self.visit(node.key), self.visit(node.value))
        self.context.pop()

        return result_type

    def visit_SetComp(self, node):
        if len(node.generators) != 1:
            raise CompileError("only one for-loop allowed", node)

        generator = node.generators[0]
        iter_type = self.visit(generator.iter)

        if isinstance(iter_type, list):
            item_type = list(iter_type)[0]
        else:
            raise CompileError(f"unsupported type '{iter_type}'", node)

        self.context.push()
        self.visit_define_local_variables(generator.target, item_type)
        result_type = Set(self.visit(node.elt))
        self.context.pop()

        return result_type



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""
Functions to manipulate packed binary representations of number sets.

To save space, coverage stores sets of line numbers in SQLite using a packed
binary representation called a numbits.  A numbits is a set of positive
integers.

A numbits is stored as a blob in the database.  The exact meaning of the bytes
in the blobs should be considered an implementation detail that might change in
the future.  Use these functions to work with those binary blobs of data.

"""


def nums_to_numbits(nums):
    """Convert `nums` into a numbits.

    Arguments:
        nums: a reusable iterable of integers, the line numbers to store.

    Returns:
        A binary blob.
    """
    try:
        nbytes = max(nums) // 8 + 1
    except ValueError:
        # nums was empty.
        return b''
    bb = bytearray(nbytes)
    for num in nums:
        bb[num//8] |= 1 << num % 8
    return bytes(bb)


def numbits_to_nums(numbits):
    """Convert a numbits into a list of numbers.

    Arguments:
        numbits: a binary blob, the packed number set.

    Returns:
        A list of ints.

    When registered as a SQLite function by :func:`register_sqlite_functions`,
    this returns a string, a JSON-encoded list of ints.

    """
    nums = []
    for byte_i, byte in enumerate(numbits):
        for bit_i in range(8):
            if (byte & (1 << bit_i)):
                nums.append(byte_i * 8 + bit_i)
    return nums



from .utils import TestCase
from .utils import build_and_test_module


class Test(TestCase):

    def test_enums(self):
        build_and_test_module('enums')

    def test_invalid_string_enum_member_value(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    A = "s"\n',
            '  File "", line 3\n'
            '        A = "s"\n'
            '            ^\n'
            "CompileError: invalid enum member value\n")

    def test_invalid_enum_member_name(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    V1, V2 = 1\n',
            '  File "", line 3\n'
            '        V1, V2 = 1\n'
            '        ^\n'
            "CompileError: invalid enum member syntax\n")

    def test_invalid_enum_member_value_plus_sign(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    A = +1\n',
            '  File "", line 3\n'
            '        A = +1\n'
            '            ^\n'
            "CompileError: invalid enum member value\n")

    def test_invalid_enum_member_value_variable(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    A = b\n',
            '  File "", line 3\n'
            '        A = b\n'
            '            ^\n'
            "CompileError: invalid enum member value\n")

    def test_non_pascal_case_enum_member_name(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    aB = 1\n',
            '  File "", line 3\n'
            '        aB = 1\n'
            '        ^\n'
            "CompileError: enum member names must be pascal case\n")

    def test_invalid_enum_member_syntax(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    1 + 1\n',
            '  File "", line 3\n'
            '        1 + 1\n'
            '        ^\n'
            "CompileError: invalid enum member syntax\n")

    def test_empty_enum_type(self):
        self.assert_transpile_raises(
            '@enum()\n'
            'class Foo:\n'
            '    Ab = 1\n',
            '  File "", line 1\n'
            '    @enum()\n'
            '     ^\n'
            "CompileError: one parameter expected, got 0\n")

    def test_bad_enum_type_f32(self):
        self.assert_transpile_raises(
            '@enum(f32)\n'
            'class Foo:\n'
            '    Ab = 1\n',
            '  File "", line 1\n'
            '    @enum(f32)\n'
            '          ^\n'
            "CompileError: integer type expected, not 'f32'\n")

    def test_enum_float_value(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    A = 1\n'
            'def foo():\n'
            '    print(Foo(0.0))\n',
            '  File "", line 5\n'
            '        print(Foo(0.0))\n'
            '                  ^\n'
            "CompileError: cannot convert float to 'i64'\n")

    def test_enum_too_many_parameters(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    A = 1\n'
            'def foo():\n'
            '    print(Foo(1, 2))\n',
            '  File "", line 5\n'
            '        print(Foo(1, 2))\n'
            '              ^\n'
            "CompileError: expected 1 parameter, got 2\n")

    def test_not_enum(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    A = 1\n'
            'def foo():\n'
            '    print(not Foo.A)\n',
            '  File "", line 5\n'
            '        print(not Foo.A)\n'
            '                  ^\n'
            "CompileError: expected a 'bool', got a 'foo.lib.Foo'\n")

    def test_enum_member_value_lower_than_previous_1(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    A = 0\n'
            '    B = -1\n',
            '  File "", line 4\n'
            '        B = -1\n'
            '            ^\n'
            "CompileError: enum member value lower than for previous member\n")

    def test_enum_member_value_lower_than_previous_2(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    A\n'
            '    B\n'
            '    C = 0\n',
            '  File "", line 5\n'
            '        C = 0\n'
            '            ^\n'
            "CompileError: enum member value lower than for previous member\n")

    def test_enum_pascal_case(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class foo:\n'
            '    A\n',
            '  File "", line 2\n'
            '    class foo:\n'
            '    ^\n'
            "CompileError: enum names must be pascal case\n")

    def test_enum_bad_member_syntax(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    def a(self):\n'
            '        pass\n',
            '  File "", line 3\n'
            '        def a(self):\n'
            '        ^\n'
            "CompileError: invalid enum member syntax\n")

    def test_use_missing_enum_value_in_print(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    Apa = 1\n'
            'def foo():\n'
            '    print(Foo.APA)\n',
            '  File "", line 5\n'
            '        print(Foo.APA)\n'
            '              ^\n'
            "CompileError: enum has no member 'APA'\n")

    def test_use_missing_enum_value_in_comparision(self):
        self.assert_transpile_raises(
            '@enum\n'
            'class Foo:\n'
            '    Apa = 1\n'
            'def foo():\n'
            '    if Foo.APA == Foo.Apa:\n'
            '        pass\n',
            '  File "", line 5\n'
            '        if Foo.APA == Foo.Apa:\n'
            '           ^\n'
            "CompileError: enum has no member 'APA'\n")



from .utils import TestCase
from .utils import build_and_test_module


class Test(TestCase):

    def test_asserts(self):
        build_and_test_module('asserts')

    def test_assert_function_that_does_not_return_any_value(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    pass\n'
            'def bar():\n'
            '    assert foo()\n',
            '  File "", line 4\n'
            '        assert foo()\n'
            '               ^\n'
            "CompileError: expected a 'bool', got 'None'\n")



import os
from textwrap import dedent
from textwrap import indent
from textwrap import wrap

from docutils import nodes
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import PythonLexer
from sphinx.directives import SphinxDirective
from sphinx.locale import _

from ..parser import ast
from ..transpiler.definitions import find_definitions
from ..transpiler.utils import METHOD_OPERATORS
from ..transpiler.utils import format_mys_type
from ..transpiler.utils import has_docstring
from ..transpiler.utils import is_private
from ..version import __version__


def is_private_method(name):
    if name in METHOD_OPERATORS:
        return False
    elif name == '__init__':
        return False
    else:
        return is_private(name)


class MysFileDirective(SphinxDirective):
    required_arguments = 1
    has_content = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = []

    def run(self):
        mys_file_path = os.path.join('..', self.arguments[0])

        with open(mys_file_path, 'r') as fin:
            source = fin.read()

        tree = ast.parse(source)
        definitions = find_definitions(tree,
                                       source.splitlines(),
                                       mys_file_path[:-4].split('/'),
                                       'lib')
        self.process_enums(definitions)
        self.process_traits(definitions)
        self.process_classes(definitions)
        self.process_functions(definitions)
        self.process_variables(definitions)

        return self.items

    def process_enums(self, definitions):
        for enum in definitions.enums.values():
            if is_private(enum.name):
                continue

            text = f'@enum({enum.type})\n'
            text += f'class {enum.name}:\n'

            if enum.docstring is not None:
                text += self.process_docstring(enum.docstring, 4)
                text += '\n\n'

            for member_name, _ in enum.members:
                text += f'    {member_name}'
                text += '\n'

            self.items.append(self.make_node(text))

    def process_docstring(self, docstring, indention):
        if docstring is None:
            return ''

        lines = docstring.splitlines()
        text = '"""'
        text += indent(lines[0] + '\n' + dedent('\n'.join(lines[1:])),
                       ' ' * indention).rstrip()
        text += '"""'

        return text

    def make_node(self, text):
        text = highlight(text, PythonLexer(), HtmlFormatter())

        return nodes.raw('',
                         text.replace('&quot;&quot;&quot;', ''),
                         format='html')

    def process_classes(self, definitions):
        for klass in definitions.classes.values():
            if is_private(klass.name):
                continue

            bases = ', '.join(klass.implements)

            if bases:
                bases = f'({bases})'

            text = f'class {klass.name}{bases}:\n'
            public_members = [member
                              for member in klass.members.values()
                              if not is_private(member.name)]

            if klass.docstring is not None:
                text += self.process_docstring(klass.docstring, 4)
                text += '\n'

                if len(public_members) > 0:
                    text += '\n'

            for member in public_members:
                text += f'    {member.name}: {format_mys_type(member.type)}\n'

            for methods in klass.methods.values():
                for method in methods:
                    if is_private_method(method.name):
                        continue

                    text += '\n'

                    if method.raises:
                        raises = ', '.join(method.raises)
                        text += f'    @raises({raises})\n'

                    signature_string = method.signature_string(True)
                    text += indent(f'def {signature_string}:', '    ')
                    text += '\n'
                    text += self.process_docstring(method.docstring, 8)
                    text = text.strip()
                    text += '\n'

            self.items.append(self.make_node(text))

    def process_traits(self, definitions):
        for trait in definitions.traits.values():
            if is_private(trait.name):
                continue

            text = '@trait\n'
            text += f'class {trait.name}:\n'

            if trait.docstring is not None:
                text += self.process_docstring(trait.docstring, 4)
                text += '\n'

            for methods in trait.methods.values():
                for method in methods:
                    if is_private_method(method.name):
                        continue

                    text += '\n'
                    text += f'    def {method.signature_string(True)}:'
                    text += '\n'
                    text += self.process_docstring(method.docstring, 8)
                    text = text.strip()
                    text += '\n'

            self.items.append(self.make_node(text))

    def process_functions(self, definitions):
        for functions in definitions.functions.values():
            for function in functions:
                if function.is_test:
                    continue

                if is_private(function.name):
                    continue

                for name in function.raises:
                    text += f'@raises({name})\n'

                text = f'def {function.signature_string(False)}:'
                text += '\n'
                text += self.process_docstring(function.docstring, 4)
                self.items.append(self.make_node(text))

    def process_variables(self, definitions):
        for variable in definitions.variables.values():
            if is_private(variable.name):
                continue

            text = f'{variable.name}: {format_mys_type(variable.type)}\n'
            text += self.process_docstring(variable.docstring, 0)
            self.items.append(self.make_node(text))

def setup(app):
    app.add_directive('mysfile', MysFileDirective)

    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True
    }
